/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "access.h"
#include "dsm.h"
#include "selectA.h"
#include "registers.h"
#include "bsp.h"
#include "datamodel.h"
#include "button.h"

Q_DEFINE_THIS_MODULE(access)

/* FSM state-handler functions */
static QState Access_initial(Access *me, QEvent const *e);

static QState Access_top(Access *me, QEvent const *e);

static QState Access_idle(Access *me, QEvent const *e);
static QState Access_leave_hold(Access *me, QEvent const *e);

/* timeout for ending hold */
#define ACCESS_HOLD_END_TIMEOUT_TICKS       (2100/BSP_MS_PER_TICKS)

/* timeout for going from present to absent */
#define ACCESS_ABSENT_TIMEOUT_TICKS         (1700/BSP_MS_PER_TICKS)

/*..........................................................................*/
void Access_ctor(Access *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&Access_initial);
}

/* HSM definition */

/*..........................................................................*/
static QState Access_initial(Access *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_FUN_DICTIONARY(&Access_top);

    QS_FUN_DICTIONARY(&Access_idle);
    QS_FUN_DICTIONARY(&Access_leave_hold);

    return Q_TRAN(&Access_idle);
}

/*..........................................................................*/
static QState Access_top(Access *me, QEvent const *e)
{
    /* implements default access behavior */
    switch (e->sig) {
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Fid_IsAccessLeaveButton(me->fid, me->key) && (SelectA_SecurityAlarmScene() == FIRE)) {
            /* call undo scene fire on all devices in all zones */
            SelectA_UndoSceneNumberBroadcast(me->localZoneId, me->groupId, FIRE, me->deviceId);
        }
        switch (FidGetSubclass(me->fid)) {
        case FID_SUBCLASS_ACCESS_LEAVING: {
            if (me->absent) {
                /* call present scene on all devices in all zones */
                SelectA_CallSceneZoneGroup(0, 0, me->localZoneId, me->groupId, PRESENT, me->deviceId);
            } else {
                /* send general acknowledge */
                SelectA_SendGeneralAck();
            }
            return Q_HANDLED();
        }
        }
        /* fall through */
    }
    case CT_TIP_2X_SIG:
        /* fall through */
    case CT_TIP_3X_SIG:
        /* fall through */
    case CT_TIP_4X_SIG:
        /* fall through */
    case CT_CLICK_2X_SIG:
        /* fall through */
    case CT_CLICK_3X_SIG:
        /* fall through */
    case CT_HOLD_START_SIG: {
        switch (FidGetSubclass(me->fid)) {
        case FID_SUBCLASS_ACCESS_BELL: {
            /* call SIG_BELL on all devices in all zones */
            SelectA_CallSceneZoneGroup(0, 0, me->localZoneId, me->groupId, SIG_BELL, me->deviceId);
            return Q_HANDLED();
        }
        }
        /* fall through */
    }
    case CT_HOLD_END_SIG: {
        /* send general acknowledge for any not handled click/tip */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Access_idle(Access *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_HOLD_START_SIG: {
        if (FidGetSubclass(me->fid) == FID_SUBCLASS_ACCESS_LEAVING) {
            /* send general acknowledge */
            SelectA_SendGeneralAck();
            if (Fid_IsAccessLeaveButton(me->fid, me->key)) {
                return Q_TRAN(&Access_leave_hold);
            }
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&Access_top);
}

/*..........................................................................*/
static QState Access_leave_hold(Access *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* arm timers */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, ACCESS_ABSENT_TIMEOUT_TICKS, ABSENT_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, ACCESS_HOLD_END_TIMEOUT_TICKS, END_ACCESS_SIG, BUTTON_ZONE_0, me->localZoneId, me->groupId, me->deviceId, FALSE, 0);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, ACCESS_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_ACCESS_SIG: {
        return Q_TRAN(&Access_idle);
    }
    case ABSENT_SIG: {
        /* call absent scene on all devices in all zones */
        SelectA_CallSceneZoneGroup(0, 0, me->localZoneId, me->groupId, ABSENT, me->deviceId);
        return Q_TRAN(&Access_idle);
    }
    }
    return Q_SUPER(&Access_top);
}
