/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM, imported from libdSvDC.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_AVAHI

#include <stdlib.h>
#include <syslog.h>
#include <avahi-common/alternative.h>

#include "discovery.h"
#include "util.h"
#include "vdc-api-proto.h"

static void vdsm_avahi_create_services(vdsm_avahi_handle_t *handle,
                                       AvahiClient *client);

static void vdsm_avahi_entry_group_callback(AvahiEntryGroup *group,
                                             AvahiEntryGroupState state,
                                             AVAHI_GCC_UNUSED void *userdata)
{
    if (userdata == NULL)
    {
        vdsm_report(LOG_ERR, "invalid userdata / missing handle\n");
        return;
    }

    vdsm_avahi_handle_t *handle = (vdsm_avahi_handle_t *)userdata;

    if (!(handle->group == group || group == NULL))
    {
        return; /* this was an assert in the avahi example */
    }

    switch (state)
    {
        case AVAHI_ENTRY_GROUP_ESTABLISHED:
            vdsm_report(LOG_DEBUG, "service \"%s\" successfully established.\n",
                   handle->name);
            break;

        case AVAHI_ENTRY_GROUP_COLLISION:
        {
            char *new_name;

            new_name = avahi_alternative_service_name(handle->name);
            if (handle->name)
            {
                avahi_free(handle->name);
            }
            handle->name = new_name;

            vdsm_report(LOG_WARNING, "service name collision, renaming service "
                   "to '%s'\n", handle->name);
            /* recreate the service */
            vdsm_avahi_create_services(handle,
                                       avahi_entry_group_get_client(group));
            break;
        }
        case AVAHI_ENTRY_GROUP_FAILURE :
            vdsm_report(LOG_ERR, "entry group failure: %s\n",
                avahi_strerror(
                    avahi_client_errno(avahi_entry_group_get_client(group))));
            /* failure while registering service */
            avahi_simple_poll_quit(handle->simple_poll);
            break;

        case AVAHI_ENTRY_GROUP_UNCOMMITED:
        case AVAHI_ENTRY_GROUP_REGISTERING:
        default:
            break;
    }
}

static void vdsm_avahi_create_services(vdsm_avahi_handle_t *handle,
                                       AvahiClient *client)
{
    int ret = 0;
    char *new_name = NULL;

    if (!client)
    {
        return;
    }

    if (!handle)
    {
        return;
    }

    /* If this is the first time we're called, let's create a new
     * entry group if necessary */

    if (!handle->group)
    {
        handle->group = avahi_entry_group_new(client,
                                    vdsm_avahi_entry_group_callback, handle);
        if (!handle->group)
        {
            vdsm_report(LOG_ERR, "avahi_entry_group_new() failed: %s\n",
                                    avahi_strerror(avahi_client_errno(client)));
            goto fail;
        }
    }
    /* add our entries if the group is empty */
    if (avahi_entry_group_is_empty(handle->group))
    {
        vdsm_report(LOG_DEBUG, "Adding service '%s'\n", handle->name);

        const char *key = "dSUID=";
        size_t len = strlen(key) + strlen(handle->vdsm_dsuid) + 1;
        char *txt = malloc(len);
        if (txt)
        {
            snprintf(txt, len, "%s%s", key, handle->vdsm_dsuid);
        }

        ret = avahi_entry_group_add_service(handle->group,
                                            AVAHI_IF_UNSPEC,
                                            AVAHI_PROTO_INET, 0,
                                            handle->name,
                                            "_ds-vdsm._tcp", NULL,
                                            NULL, handle->vdsm_port, txt,
                                            NULL);
        if (txt)
        {
            free(txt);
        }

        if (ret < 0)
        {
            if (ret == AVAHI_ERR_COLLISION)
            {
                goto collision;
            }

            vdsm_report(LOG_ERR, "could not add _ds-vdc._tcp service: %s\n",
                   avahi_strerror(ret));
            goto fail;
        }

        /* the service */
        ret = avahi_entry_group_commit(handle->group);
        if (ret < 0)
        {
            vdsm_report(LOG_ERR, "could not commit entry group: %s\n",
                   avahi_strerror(ret));
            goto fail;
        }
    }

    return;

collision:

    /* service name collision with a local service happened, pick a new name */
    new_name = avahi_alternative_service_name(handle->name);
    if (handle->name)
    {
        avahi_free(handle->name);
    }
    handle->name = new_name;

    vdsm_report(LOG_WARNING, "service name collision, renaming service to '%s'\n",
           handle->name);

    avahi_entry_group_reset(handle->group);

    vdsm_avahi_create_services(handle, client);
    return;

fail:
    avahi_simple_poll_quit(handle->simple_poll);
}

void vdsm_avahi_resolve_callback(
        AvahiServiceResolver *resolver,
        AVAHI_GCC_UNUSED AvahiIfIndex interface,
        AVAHI_GCC_UNUSED AvahiProtocol protocol,
        AvahiResolverEvent event,
        const char *name,
        const char *type,
        const char *domain,
        const char *host_name,
        const AvahiAddress *address,
        uint16_t port,
        AVAHI_GCC_UNUSED AvahiStringList *txt,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        AVAHI_GCC_UNUSED void* userdata)
{
    if (!resolver) {
        return;
    }

    /* callback is called whenever a service has been resolved successfully or
       timed out */

    switch (event)
    {
        case AVAHI_RESOLVER_FAILURE:
            vdsm_report(LOG_ERR, "failed to resolve service \"%s\" of type \"%s\" "
                  "in domain \"%s\": %s\n", name, type, domain,
                  avahi_strerror(avahi_client_errno(
                          avahi_service_resolver_get_client(resolver))));
            break;
        case AVAHI_RESOLVER_FOUND:
        {
            char ip[AVAHI_ADDRESS_STR_MAX];
            avahi_address_snprint(ip, sizeof(ip), address);
            char *dsuid = NULL;

            AvahiStringList *pair = avahi_string_list_find(txt, "dSUID");
            if (pair) {
                dsuid = strchr(avahi_string_list_get_text(pair), '=');
                if (dsuid) {
                    // skip '='
                    dsuid++;
                }
            }

            if (!dsuid) {
                vdsm_report(LOG_DEBUG, "ignoring vDC %s at: %s:%u (%s) - missing "
                        "dSUID in announcement\n", name, host_name, port, ip);
                break;

            }
            vdsm_report(LOG_DEBUG, "found %s at: %s:%u (%s)\n", name, host_name,
                   port, ip);
            char str_port[6]; /* "65535\0" */
            snprintf(str_port, 6, "%u", port);
            vdc_api_add_auto_connection(host_name, str_port, dsuid, name);
            break;
        }
        default:
            break;
    }

    avahi_service_resolver_free(resolver);
}

void vdsm_avahi_browse_callback(
        AvahiServiceBrowser *browser,
        AvahiIfIndex interface,
        AvahiProtocol protocol,
        AvahiBrowserEvent event,
        const char *name,
        const char *type,
        const char *domain,
        AVAHI_GCC_UNUSED AvahiLookupResultFlags flags,
        void* userdata)
{
    if (!browser) {
        vdsm_report(LOG_ERR, "avahi: invalid avahi service browser handle");
        return;
    }

    if (!userdata) {
        vdsm_report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    vdsm_avahi_handle_t *avahi = userdata;

    /* called when a new services becomes available or is removed from LAN */

    switch (event) {
        case AVAHI_BROWSER_FAILURE:

            vdsm_report(LOG_ERR, "avahi browse callback %s\n",
                   avahi_strerror(avahi_client_errno(
                           avahi_service_browser_get_client(browser))));
            avahi_simple_poll_quit(avahi->simple_poll);
            avahi->initialized = FALSE;
            return;

        case AVAHI_BROWSER_NEW:
            vdsm_report(LOG_DEBUG, "browser: new service \"%s\" of type \"%s\" in "
                   "domain \"%s\"\n", name, type, domain);

            /* We ignore the returned resolver object. In the callback
               function we free it. If the server is terminated before
               the callback function is called the server will free
               the resolver for us. */

            if (!(avahi_service_resolver_new(avahi->client,
                            interface, protocol, name, type, domain,
                            AVAHI_PROTO_INET, 0, vdsm_avahi_resolve_callback,
                            avahi->client)))
                vdsm_report(LOG_ERR, "failed to resolve service \"%s\": %s\n", name,
                       avahi_strerror(avahi_client_errno(avahi->client)));

            break;

        case AVAHI_BROWSER_REMOVE:
            vdsm_report(LOG_DEBUG, "browser: remove service \"%s\" of type \"%s\" "
                              "in domain '%s'\n", name, type, domain);
            vdc_api_remove_auto_connection(name);
            break;

        case AVAHI_BROWSER_ALL_FOR_NOW:
        case AVAHI_BROWSER_CACHE_EXHAUSTED:
            vdsm_report(LOG_DEBUG, "browser: %s\n",
                    event == AVAHI_BROWSER_CACHE_EXHAUSTED ?
                        "CACHE_EXHAUSTED" : "ALL_FOR_NOW");
        default:
            break;
    }
}

void vdsm_avahi_client_callback(
        AvahiClient *client,
        AvahiClientState state,
        AVAHI_GCC_UNUSED void * userdata)
{
    if (!client) {
        vdsm_report(LOG_ERR, "avahi: invalid client handle\n");
        return;
    }

    if (!userdata) {
        vdsm_report(LOG_ERR, "avahi: missing user data\n");
        return;
    }

    vdsm_avahi_handle_t *handle = userdata;

    switch (state)
    {
        case AVAHI_CLIENT_S_RUNNING:
            /* the server has startup successfully and registered its host
               name on the network, create our services */
            if (handle->announce) {
                vdsm_avahi_create_services(handle, client);
            }
            break;

        case AVAHI_CLIENT_S_COLLISION:
            /* Let's drop our registered services. When the server is back
             * in AVAHI_SERVER_RUNNING state we will register them
             * again with the new host name. */
        case AVAHI_CLIENT_S_REGISTERING:
            /* The server records are now being established. This
             * might be caused by a host name change. We need to wait
             * for our own records to register until the host name is
             * properly esatblished. */

            if (handle->group)
            {
                avahi_entry_group_reset(handle->group);
            }
            break;

        case AVAHI_CLIENT_FAILURE:
            vdsm_report(LOG_ERR, "avahi server conection failure: %s\n",
                   avahi_strerror(avahi_client_errno(client)));
            avahi_simple_poll_quit(handle->simple_poll);
            break;

        case AVAHI_CLIENT_CONNECTING:
        default:
            break;
    }
}

#endif//HAVE_AVAHI
